/*  demorse.c
 *
 *  Main body of demorse application
 */

/*
 *  demorse: An application to decode Morse signals to text
 *
 *  Copyright (C) 2002  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "demorse.h"

int
  dsp_fd, /* File descriptor of dsp device   */
  mix_fd; /* File descriptor of mixer device */

/* Signal handler */
static void sig_handler( int signal );

int
  /* Initial unit element as fragments/element */
  unit_elem = (60*AUDIO_FREQUENCY)/(50*CYCLES_PER_FRAG*INITIAL_WPM),
  /* Initial Morse speed at start up as fragments/element */
  initial_period = (60*AUDIO_FREQUENCY)/(50*CYCLES_PER_FRAG*INITIAL_WPM);

/* Runtime config data */
rc_data_t rc_data;

/*------------------------------------------------------------------------*/

  int
main( int argc, char *argv[] )
{
  int
	option,    /* Command line option returned by getopt() */
	line_idx;  /* Number of characters printed on line     */

  /* Decoded Morse character */
  char dec_char;

  /* getopt() variables */
  extern char *optarg;
  extern int optind, opterr, optopt;

  /* New and old actions for sigaction() */
  struct sigaction sa_new, sa_old;

  /* Initialize new actions */
  sa_new.sa_handler = sig_handler;
  sigemptyset( &sa_new.sa_mask );
  sa_new.sa_flags = 0;

  /* Register function to handle signals */
  sigaction( SIGINT,  &sa_new, &sa_old );
  sigaction( SIGSEGV, &sa_new, 0 );
  sigaction( SIGFPE,  &sa_new, 0 );
  sigaction( SIGTERM, &sa_new, 0 );
  sigaction( SIGABRT, &sa_new, 0 );

  /*** Process command line options ***/

  /* Initialize defaults */
  Set_Flag( ADAPT_SPEED );
  line_idx = rc_data.wrd_wrp = 0;

  strncpy( rc_data.mix_dev, MIXER_DEV,  20 ); /* Mixer device name */
  strncpy( rc_data.dsp_dev, DSP_DEV,    20 ); /* DSP device name   */
  strncpy( rc_data.rec_src, RECORD_SRC, 20 ); /* Recording/capture source */
  strncpy( rc_data.inp_lev, INPUT_LEV,  20 ); /* Input level contol name  */

  rc_data.use_chn = USE_CHANNEL;	/* Channel in use: 0=left, 1=right */
  rc_data.rec_lev = CAPTURE_LEV;	/* Recording/Capture level */
  rc_data.spd_wpm = INITIAL_WPM;	/* Morse speed WPM */
  rc_data.wrd_wrp = LINE_LENGTH;	/* Maximum line length if word wrap enabled   */
  rc_data.det_thr = DET_THRESHOLD;  /* Signal threshold for Mark/Space separation */

  /* Process command line options */
  while( (option = getopt(argc, argv, "c:d:f:l:m:p:r:s:u:w:hiv") ) != -1 )
	switch( option )
	{
	  case 'c': /* Capture level */
		rc_data.rec_lev = atoi( optarg );
		if( (rc_data.rec_lev < 0) || (rc_data.rec_lev > 100) )
		{
		  fprintf( stderr, "demorse: capture level out of range" );
		  Usage();
		  exit( -1 );
		}
		printf("demorse: setting capture level to %d\n", rc_data.rec_lev );
		break;
	  
	  case 'd': /* Detector threshold */
		rc_data.det_thr = atoi( optarg );
		if( (rc_data.det_thr < 0) || (rc_data.det_thr > 100) )
		{
		  fprintf( stderr, "demorse: detector threshold out of range" );
		  Usage();
		  exit( -1 );
		}
		printf("demorse: setting detector threshold to %d\n", rc_data.det_thr );
		break;

	  case 'f': /* Decode at fixed speed */
		unit_elem = atoi( optarg );
		Clear_Flag( ADAPT_SPEED );

		/* Speed limit 10-30 wpm */
		if( (unit_elem > MAX_UNIT_LEN) || (unit_elem < MIN_UNIT_LEN) )
		{
		  fprintf( stderr, "demorse: Morse code speed out of range" );
		  Usage();
		  exit( -1 );
		}

		/* Convert speed to frags per element */
		printf( "demorse: fixing Morse decoding speed to %d wpm\n", unit_elem );
		unit_elem = (60*AUDIO_FREQUENCY)/(50*CYCLES_PER_FRAG*unit_elem);
		break;

	  case 'l': /* Recording level device */
		strncpy( rc_data.inp_lev, optarg, 20 );
		printf( "demorse: setting recording level device to %s\n", rc_data.inp_lev );
		break;

	  case 'm': /* Mixer device */
		strncpy( rc_data.mix_dev, optarg, 20 );
		printf( "demorse: setting mixer device to %s\n", rc_data.mix_dev );
		break;

	  case 'p': /* DSP device */
		strncpy( rc_data.dsp_dev, optarg, 20 );
		printf( "demorse: setting dsp device to %s\n", rc_data.dsp_dev );
		break;

	  case 'r': /* Recording source */
		strncpy( rc_data.rec_src, optarg, 20 );
		printf( "demorse: setting recording source to %s\n", rc_data.rec_src );
		break;

	  case 's': /* Specify Morse code starting speed WPM */
		initial_period = atoi( optarg );
		Set_Flag( ADAPT_SPEED );

		/* Speed limit 10-30 wpm */
		if( (initial_period > MAX_UNIT_LEN) || (initial_period < MIN_UNIT_LEN) )
		{
		  fprintf( stderr, "%s", "demorse: Morse code speed out of range" );
		  Usage();
		  exit( -1 );
		}

		/* Convert speed to frags per element */
		printf( "demorse: setting initial Morse decoding speed to %d wpm\n",
			initial_period );
		unit_elem = initial_period =
		  (60*AUDIO_FREQUENCY)/(50*CYCLES_PER_FRAG*initial_period);
		break;

	  case 'u': /* Channel to use left/right */
		if( strcmp(optarg, "left") == 0 )
		  rc_data.use_chn = 0;
		else if( strcmp(optarg, "right") == 0 )
		  rc_data.use_chn = 1;
		else
		{
		  fprintf( stderr, "demorse: unrecognized option: %s\n", optarg );
		  Usage();
		  exit( -1 );
		}
		printf( "demorse: setting recording source to %s\n", optarg );
		break;

	  case 'w':	/* Specify word wrap line length */
		rc_data.wrd_wrp = atoi( optarg );

		/* Check word wrap line length */
		if( (rc_data.wrd_wrp > LINE_LENGTH) || (rc_data.wrd_wrp <= 0) )
		{
		  fprintf( stderr, "%s", "demorse: word wrap line length out of range" );
		  Usage();
		  exit( -1 );
		}

		Set_Flag( WORD_WRAP );
		printf( "demorse: setting word wrap to %d columns\n", rc_data.wrd_wrp );
		break;

	  case 'h': /* Print usage and exit */
		Usage();
		exit( 0 );

	  case 'i': /* Display information for signal input setup */
		Set_Flag( SETUP_LEVEL );
		break;

	  case 'v': /* Print version */
		printf( "%s\n", VERSION );
		exit( 0 );

	  default: /* Print usage and exit */
		Usage();
		exit( -1 );

	} /* End of switch( option ) */

  /* Prepare sound card */
  Setup_Sound_Card();

  /* Start audio level setup loop */
  if( isFlagSet(SETUP_LEVEL) )
	Setup_Signal_Input();

  /* Decode and print Morse signals until stopped */
  while( 1 )
  {
	/* If word wrap is enabled, wrap on 'space' char */
	if( isFlagSet(WORD_WRAP) )
	{
	  dec_char = Get_Character();
	  if( (dec_char == ' ') && (line_idx >= rc_data.wrd_wrp) )
	  {
		line_idx = 0;
		printf( "%c", '\n' );
		fflush( stdout );
	  }
	  else
	  {
		line_idx++;
		printf( "%c", dec_char );
		fflush( stdout );
	  } /* if( ((dec_char = Get_Character()) == ' ') && */

	} /* if( isFlagSet(WORD_WRAP) ) */
	else
	{
	  /* Just print chars if no word wrap */
	  printf( "%c", Get_Character() );
	  fflush( stdout );
	}

  } /* while( 1 ) */

} /* End of  main() */

/*------------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n\n",
	  "Usage: demorse [-c<capture level>][-d<detector threshold>][-f<fixed speed>]\n"
	  "               [-l<level device>][-m<mixer device>][-p<dsp device>]\n"
	  "               [-r<recording source>][-s<initial speed>][-u<channel>]\n"
	  "               [-w<line length>][-hiv]" );

  fprintf( stderr, "%s\n\n",
	  "       -c<capture level>: Capture/Recording source level.\n"
	  "       Value should be 0-100, default is 50." );

  fprintf( stderr, "%s\n\n",
	  "       -d<detector threshold>: Tone/No-tone detector threshold.\n"
	  "       Value should be 0-100, default is 60." );

  fprintf( stderr, "%s\n\n",
	  "       -f<fixed-speed>: Fixed speed in w.p.m at which\n"
	  "       to expect and to decode a Morse coded signal.\n"
	  "       Valid range is 10-30 words per minute." );

  fprintf( stderr, "%s\n\n",
	  "       -l<level device>: Input level device. Default is \"igain\"." );

  fprintf( stderr, "%s\n\n",
	  "       -m<mixer device>: Mixer device. Default is \"/dev/mixer\"." );

  fprintf( stderr, "%s\n\n",
	  "       -p<dsp device>: DSP device. Default is \"/dev/dsp\"." );

  fprintf( stderr, "%s\n\n",
	  "       -r<recording source>: Recording/Capture source. Default is \"line\"." );

  fprintf( stderr, "%s\n\n",
	  "       -s<initial speed>: Initial speed in w.p.m at start-up,\n"
	  "       at which to expect and to decode a Morse code signal.\n"
	  "       Actual decoding speed then adapts to the incoming Morse\n"
	  "       signal, limited to the range 10-30 words per minute." );

  fprintf( stderr, "%s\n\n",
	  "       -u<channel>: Channel to use. Value should be left|right.\n"
	  "       Default is \"right\"." );

  fprintf( stderr, "%s\n\n",
	  "       -w<line length>: Line length in columns beyond which\n"
	  "       word wrapping is performed. Maximum is LINE_LENGTH columns." );

  fprintf( stderr, "%s\n\n",
	  "       -h: Print this usage information and exit.");

  fprintf( stderr, "%s\n\n",
	  "       -i: Display information for setting up signal input.");

  fprintf( stderr, "%s\n\n",
	  "       -v: Print version number and exit.");

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  sig_handler()
 *
 *  Signal Action Handler function
 */

static void sig_handler( int signal )
{
  /* Wrap up and quit */
  if( dsp_fd > 0 ) close( dsp_fd );
  if( mix_fd > 0 ) close( mix_fd );

  fprintf( stderr, "\n" );
  switch( signal )
  {
	case SIGINT :
	  fprintf( stderr, "%s\n",  "demorse: Exiting via User Interrupt" );
	  exit( signal );

	case SIGSEGV :
	  fprintf( stderr, "%s\n",  "demorse: Segmentation Fault" );
	  exit( signal );

	case SIGFPE :
	  fprintf( stderr, "%s\n",  "demorse: Floating Point Exception" );
	  exit( signal );

	case SIGABRT :
	  fprintf( stderr, "%s\n",  "demorse: Abort Signal received" );
	  exit( signal );

	case SIGTERM :
	  fprintf( stderr, "%s\n",  "demorse: Termination Request received" );
	  exit( signal );
  }

} /* End of sig_handler() */

/*------------------------------------------------------------------------*/

/*  Abort_On_Error()
 *
 *  Prints an error message and exits
 */

  void
Abort_On_Error( int why )
{
  switch( why )
  {
	fprintf( stderr, "\n" );
	case -1 : /* Abort if samples buffer cannot be filled */
	  fprintf( stderr, "%s\n",
		  "demorse: Error filling signal samples buffer - aborting" );
	  break;

	case -2 : /* Abort if recording source selection fails */
	  fprintf( stderr, "%s\n",
		  "demorse: Recording source selection failed - aborting" );
	  break;

	case -3 : /* Abort sample size setting fails */
	  fprintf( stderr, "%s\n",
		  "demorse: Sample size setting failed - aborting" );
	  break;

	case -4 : /* Abort if recording speed setting fails */
	  fprintf( stderr, "%s\n",
		  "demorse: Recording speed setting failed - aborting" );
	  break;

	case -5 : /* Abort if recording level setting fails */
	  fprintf( stderr, "%s\n",
		  "demorse: Recording level setting failed - aborting" );
	  break;

	case -6 : /* Abort if samples buffer cannot be filled */
	  fprintf( stderr, "%s\n",
		  "demorse: Recording buffer not filled by read() - aborting" );
	  break;

	case -7 : /* Abort if no recording level device found */
	  fprintf( stderr, "%s\n",
		  "demorse: No recording level device found - aborting" );
	  break;

	case -8 : /* Abort if Line level setting fails */
	  fprintf( stderr, "%s\n",
		  "demorse: Line level setting failed - aborting" );
	  break;

	case -9 : /* Abort if Stereo/Mono mode setting fails */
	  fprintf( stderr, "%s\n",
		  "demorse: Stereo/Mono mode setting failed - aborting" );

  }  /* switch( why ) */

  exit( why );

} /* End of Abort_On_Error() */

/*------------------------------------------------------------------------*/

/* Functions for testing and setting/clearing flow control flags
 *
 *  See demorse.h for definition of flow control flags
 */

/* An int variable holding the single-bit flags */
static int Flags = 0;

  int
isFlagSet( int flag )
{
  return( (Flags & flag) == flag );
}

  int
isFlagClear( int flag )
{
  return( (~Flags & flag) == flag );
}

  void
Set_Flag( int flag )
{
  Flags |= flag;
}

  void
Clear_Flag( int flag )
{
  Flags &= ~flag;
}

  void
Toggle_Flag( int flag )
{
  Flags ^= flag;
}

/*------------------------------------------------------------------------*/
